@@ -9,7 +9,7 @@ namespace Mono.Debugging.Backend.Mdb
 
 		//callbacks
 		//FIXME: better naming for event callbacks
-		void OnMainProcessCreated(int process_id);
+		void NotifyStarted ();
 		
 		void OnTargetEvent (TargetEventArgs args);
 		
@@ -20,5 +20,9 @@ namespace Mono.Debugging.Backend.Mdb
 		bool OnCustomBreakpointAction (string actionId, object handle);
 		
 		void UpdateBreakpoint (object handle, int count, string lastTrace);
+		
+		void NotifySourceFileLoaded (string[] fullFilePaths);
+		
+		void NotifySourceFileUnloaded (string[] fullFilePaths);
 	}
 }
