@@ -1,4 +1,5 @@
 using System;
+using System.IO;
 using System.Collections.Generic;
 using System.Runtime.InteropServices;
 using System.Runtime.Remoting.Lifetime;
@@ -13,6 +14,7 @@ using DL = Mono.Debugging.Client;
 using DB = Mono.Debugging.Backend;
 
 using Mono.Debugging.Client;
+using Mono.Debugging.Evaluation;
 using Mono.Debugging.Backend.Mdb;
 
 namespace DebuggerServer
@@ -24,13 +26,13 @@ namespace DebuggerServer
 		MD.DebuggerSession session;
 		MD.Process process;
 		MD.GUIManager guiManager;
+		MdbAdaptor mdbAdaptor;
 		int max_frames;
-		bool internalInterruptionRequested;
 		List<ST.WaitCallback> stoppedWorkQueue = new List<ST.WaitCallback> ();
 		List<ST.WaitCallback> eventQueue = new List<ST.WaitCallback> ();
 		bool initializing;
 		bool running;
-		ExpressionEvaluator evaluator = new NRefactoryEvaluator ();
+		NRefactoryEvaluator evaluator = new NRefactoryEvaluator ();
 		MD.Thread activeThread;
 		Dictionary<int,BreakEvent> events = new Dictionary<int,BreakEvent> ();
 		Dictionary<int,string> lastConditionValue = new Dictionary<int,string> ();
@@ -39,29 +41,26 @@ namespace DebuggerServer
 		Dictionary<int, ST.WaitCallback> breakUpdates = new Dictionary<int,ST.WaitCallback> ();
 		bool breakUpdateEventsQueued;
 		
-		AsyncEvaluationTracker asyncEvaluationTracker;
-		RuntimeInvokeManager invokeManager;
-
+		MdbObjectValueAdaptor mdbObjectValueAdaptor;
+		
 		public const int DefaultAsyncSwitchTimeout = 60;
 		public const int DefaultEvaluationTimeout = 1000;
 		public const int DefaultChildEvaluationTimeout = 5000;
 		
 		const int BreakEventUpdateNotifyDelay = 500;
-
+		
 		public DebuggerServer (IDebuggerController dc)
 		{
 			this.controller = dc;
+			mdbObjectValueAdaptor = new MdbObjectValueAdaptor ();
 			MarshalByRefObject mbr = (MarshalByRefObject)controller;
 			ILease lease = mbr.GetLifetimeService() as ILease;
 			lease.Register(this);
 			max_frames = 100;
-			
+				
 			ST.Thread t = new ST.Thread ((ST.ThreadStart)EventDispatcher);
 			t.IsBackground = true;
 			t.Start ();
-
-			invokeManager = new RuntimeInvokeManager ();
-			asyncEvaluationTracker = new AsyncEvaluationTracker ();
 		}
 		
 		public override object InitializeLifetimeService ()
@@ -69,28 +68,69 @@ namespace DebuggerServer
 			return null;
 		}
 		
-		public ExpressionEvaluator Evaluator {
+		public MdbObjectValueAdaptor MdbObjectValueAdaptor {
+			get { return mdbObjectValueAdaptor; }
+		}
+		
+		public Mono.Debugging.Evaluation.ExpressionEvaluator Evaluator {
 			get { return evaluator; }
 		}
 
-		public AsyncEvaluationTracker AsyncEvaluationTracker {
-			get {
-				return asyncEvaluationTracker;
-			}
+		public MdbAdaptor MdbAdaptor {
+			get { return mdbAdaptor; }
 		}
 
 		#region IDebugger Members
+		
+		public string InitializeMdb (string mdbVersion, int mdPid)
+		{
+			mdbAdaptor = MdbAdaptorFactory.CreateAdaptor (mdbVersion);
+			WatchParentProcess (mdPid);
+			return mdbAdaptor.MdbVersion;
+		}
+	
+		void WatchParentProcess (int pid)
+		{
+			ST.Thread t = new ST.Thread (delegate () {
+				while (true) {
+					try {
+						System.Diagnostics.Process.GetProcessById (pid);
+					} catch {
+						ForceShutdown ();
+					}
+					ST.Thread.Sleep (1000);
+				}
+			});
+			t.IsBackground = true;
+			t.Start ();
+		}
+		
+		void ForceShutdown ()
+		{
+			try {
+				Exit ();
+			} catch {
+			}
+			Environment.Exit (1);
+		}
 
-		public void Run (DL.DebuggerStartInfo startInfo)
+		public void Run (MonoDebuggerStartInfo startInfo)
 		{
 			try {
 				if (startInfo == null)
 					throw new ArgumentNullException ("startInfo");
+			
+				Console.WriteLine ("MDB version: " + mdbAdaptor.MdbVersion);
+				
+				mdbAdaptor.StartInfo = startInfo;
 
 				Report.Initialize ();
 
 				DebuggerConfiguration config = new DebuggerConfiguration ();
 				config.LoadConfiguration ();
+				mdbAdaptor.Configuration = config;
+				mdbAdaptor.InitializeConfiguration ();
+				
 				debugger = new MD.Debugger (config);
 				
 				debugger.ModuleLoadedEvent += OnModuleLoadedEvent;
@@ -98,24 +138,35 @@ namespace DebuggerServer
 				
 				debugger.ProcessReachedMainEvent += delegate (MD.Debugger deb, MD.Process proc) {
 					OnInitialized (deb, proc);
-					NotifyStarted ();
 				};
 
+				if (startInfo.IsXsp) {
+					mdbAdaptor.SetupXsp ();
+					config.FollowFork = false;
+				}
+				config.OpaqueFileNames = false;
+				
 				DebuggerOptions options = DebuggerOptions.ParseCommandLine (new string[] { startInfo.Command } );
 				options.WorkingDirectory = startInfo.WorkingDirectory;
 				Environment.CurrentDirectory = startInfo.WorkingDirectory;
 				options.StopInMain = false;
 				
 				if (!string.IsNullOrEmpty (startInfo.Arguments))
-					options.InferiorArgs = startInfo.Arguments.Split(' ');
+					options.InferiorArgs = ToArgsArray (startInfo.Arguments);
 				
 				if (startInfo.EnvironmentVariables != null) {
 					foreach (KeyValuePair<string,string> env in startInfo.EnvironmentVariables)
 						options.SetEnvironment (env.Key, env.Value);
 				}
 				session = new MD.DebuggerSession (config, options, "main", null);
+				mdbAdaptor.Session = session;
+				mdbAdaptor.InitializeSession ();
 				
-				debugger.Run(session);
+				ST.ThreadPool.QueueUserWorkItem (delegate {
+					// Run in a thread to avoid a deadlock, since NotifyStarted calls back to the client.
+					NotifyStarted ();
+					debugger.Run(session);
+				});
 
 			} catch (Exception e) {
 				Console.WriteLine ("error: " + e.ToString ());
@@ -123,19 +174,22 @@ namespace DebuggerServer
 			}
 		}
 		
-		public void AttachToProcess (int pid)
+		public void AttachToProcess (long pid)
 		{
 			Report.Initialize ();
 
 			DebuggerConfiguration config = new DebuggerConfiguration ();
+			mdbAdaptor.Configuration = config;
+			mdbAdaptor.InitializeConfiguration ();
 			config.LoadConfiguration ();
 			debugger = new MD.Debugger (config);
 
 			DebuggerOptions options = DebuggerOptions.ParseCommandLine (new string[0]);
 			options.StopInMain = false;
 			session = new MD.DebuggerSession (config, options, "main", (IExpressionParser) null);
+			mdbAdaptor.Session = session;
 			
-			Process proc = debugger.Attach (session, pid);
+			Process proc = debugger.Attach (session, (int)pid);
 			OnInitialized (debugger, proc);
 			
 			ST.ThreadPool.QueueUserWorkItem (delegate {
@@ -146,28 +200,20 @@ namespace DebuggerServer
 		public void Detach ()
 		{
 			CancelRuntimeInvokes ();
-			RunWhenStopped (delegate {
-				try {
-					debugger.Detach ();
-				} catch (Exception ex) {
-					Console.WriteLine (ex);
-				} finally {
-					running = false;
-				}
-			});
+			try {
+				debugger.Detach ();
+			} catch (Exception ex) {
+				Console.WriteLine (ex);
+			} finally {
+				running = false;
+			}
 		}
 
 		public void Stop ()
 		{
 			CancelRuntimeInvokes ();
 			QueueTask (delegate {
-				if (internalInterruptionRequested) {
-					// Stop already internally requested. By resetting the flag, the interruption
-					// won't be considered internal anymore and the session won't be automatically restarted.
-					internalInterruptionRequested = false;
-				}
-				else
-					guiManager.Stop (process.MainThread);
+				guiManager.Stop (process.MainThread);
 			});
 		}
 
@@ -179,7 +225,6 @@ namespace DebuggerServer
 			running = false;
 		}
 
-		int ncc = 0;
 		public void NextLine ()
 		{
 			if (running)
@@ -229,6 +274,12 @@ namespace DebuggerServer
 				guiManager.Continue (activeThread);
 			});
 		}
+		
+		public bool AllowBreakpointChanges {
+			get {
+				return mdbAdaptor == null || mdbAdaptor.AllowBreakEventChanges;
+			}
+		}
 
 		public int InsertBreakEvent (DL.BreakEvent be, bool enable)
 		{
@@ -238,7 +289,10 @@ namespace DebuggerServer
 			
 			if (bp != null) {
 				MD.SourceLocation location = new MD.SourceLocation (bp.FileName, bp.Line);
-				ev = session.InsertBreakpoint (ThreadGroup.Global, location);
+				MD.SourceBreakpoint sbp = new MD.SourceBreakpoint (session, ThreadGroup.Global, location);
+				mdbAdaptor.InitializeBreakpoint (sbp);
+				session.AddEvent (sbp);
+				ev = sbp;
 			}
 			else if (be is Catchpoint) {
 				Catchpoint cp = (Catchpoint) be;
@@ -256,18 +310,14 @@ namespace DebuggerServer
 			ev.IsEnabled = enable;
 			
 			if (!initializing) {
-				RunWhenStopped (delegate {
-					try {
-						ev.Activate (process.MainThread);
-					} catch (Exception ex) {
-						Console.WriteLine (ex);
-					}
-				});
+				lock (debugger) {
+					mdbAdaptor.ActivateEvent (ev);
+				}
 			}
 						                    
-			if (bp != null && !running && activeThread.CurrentFrame != null && !string.IsNullOrEmpty (bp.ConditionExpression) && bp.BreakIfConditionChanges) {
+			if (bp != null && !running && !initializing && activeThread.CurrentFrame != null && !string.IsNullOrEmpty (bp.ConditionExpression) && bp.BreakIfConditionChanges) {
 				// Initial expression evaluation
-				EvaluationContext ctx = new EvaluationContext (activeThread, activeThread.CurrentFrame, -1);
+				MdbEvaluationContext ctx = new MdbEvaluationContext (activeThread, activeThread.CurrentFrame, -1);
 				ML.TargetObject ob = EvaluateExp (ctx, bp.ConditionExpression);
 				if (ob != null)
 					lastConditionValue [ev.Index] = evaluator.TargetObjectToExpression (ctx, ob);
@@ -280,27 +330,18 @@ namespace DebuggerServer
 		public void RemoveBreakEvent (int handle)
 		{
 			CancelRuntimeInvokes ();
-			RunWhenStopped (delegate {
-				try {
-					Event ev = session.GetEvent (handle);
-					session.DeleteEvent (ev);
-					events.Remove (handle);
-				} catch (Exception ex) {
-					Console.WriteLine (ex);
-				}
-			});
+			Event ev = session.GetEvent (handle);
+			mdbAdaptor.RemoveEvent (ev);
 		}
 		
 		public void EnableBreakEvent (int handle, bool enable)
 		{
 			CancelRuntimeInvokes ();
-			RunWhenStopped (delegate {
-				Event ev = session.GetEvent (handle);
-				if (enable)
-					ev.Activate (process.MainThread);
-				else
-					ev.Deactivate (process.MainThread);
-			});
+			Event ev = session.GetEvent (handle);
+			if (enable)
+				ev.Activate (process.MainThread);
+			else
+				ev.Deactivate (process.MainThread);
 		}
 		
 		public object UpdateBreakEvent (object handle, DL.BreakEvent bp)
@@ -330,7 +371,7 @@ namespace DebuggerServer
 				return false;
 			}
 
-			EvaluationContext ctx = new EvaluationContext (frame.Thread, frame, -1);
+			MdbEvaluationContext ctx = new MdbEvaluationContext (frame.Thread, frame, -1);
 			DL.Breakpoint bp = be as DL.Breakpoint;
 			if (bp != null && !string.IsNullOrEmpty (bp.ConditionExpression)) {
 				ML.TargetObject val = EvaluateExp (ctx, bp.ConditionExpression);
@@ -381,7 +422,7 @@ namespace DebuggerServer
 				EvaluationOptions ops = new EvaluationOptions ();
 				ops.CanEvaluateMethods = true;
 				var = (ValueReference) Server.Instance.Evaluator.Evaluate (ctx, exp, ops);
-				return var.Value;
+				return (ML.TargetObject) var.Value;
 			} catch {
 				return null;
 			}
@@ -425,7 +466,7 @@ namespace DebuggerServer
 				controller.UpdateBreakpoint (eventHandle, hitCount, lastTrace);
 		}
 
-		public ThreadInfo[] GetThreads (int processId)
+		public ThreadInfo[] GetThreads (long processId)
 		{
 			MD.Process p = GetProcess (processId);
 			if (p == null)
@@ -457,7 +498,7 @@ namespace DebuggerServer
 			return new ThreadInfo (t.Process.ID, t.ID, t.Name, loc);
 		}
 		
-		public DL.Backtrace GetThreadBacktrace (int processId, int threadId)
+		public DL.Backtrace GetThreadBacktrace (long processId, long threadId)
 		{
 			MD.Thread t = GetThread (processId, threadId);
 			if (t != null && t.IsStopped)
@@ -466,27 +507,27 @@ namespace DebuggerServer
 				return null;
 		}
 		
-		public void SetActiveThread (int processId, int threadId)
+		public void SetActiveThread (long processId, long threadId)
 		{
 			activeThread = GetThread (processId, threadId);
 		}
 
-		MD.Thread GetThread (int procId, int threadId)
+		MD.Thread GetThread (long procId, long threadId)
 		{
 			MD.Process proc = GetProcess (procId);
 			if (proc != null) {
 				foreach (MD.Thread t in proc.GetThreads ()) {
-					if (t.ID == threadId)
+					if (t.ID == (int)threadId)
 						return t;
 				}
 			}
 			return null;
 		}
 		
-		MD.Process GetProcess (int id)
+		MD.Process GetProcess (long id)
 		{
 			foreach (MD.Process p in debugger.Processes) {
-				if (p.ID == id)
+				if (p.ID == (int)id)
 					return p;
 			}
 			return null;
@@ -548,11 +589,13 @@ namespace DebuggerServer
 				Console.WriteLine (ex);
 		}
 		
-		public ML.TargetObject RuntimeInvoke (EvaluationContext ctx, ML.TargetFunctionType function,
+		public ML.TargetObject RuntimeInvoke (MdbEvaluationContext ctx, ML.TargetFunctionType function,
 							  ML.TargetStructObject object_argument,
 							  params ML.TargetObject[] param_objects)
 		{
-			return invokeManager.Invoke (ctx, function, object_argument, param_objects);
+			MethodCall mc = new MethodCall (ctx, function, object_argument, param_objects);
+			ctx.Adapter.AsyncExecute (mc, ctx.Timeout);
+			return mc.ReturnValue;
 		}
 
 		DL.Backtrace CreateBacktrace (MD.Thread thread)
@@ -599,6 +642,8 @@ namespace DebuggerServer
 			this.process = process;
 			this.debugger = debugger;
 			
+			mdbAdaptor.Process = process;
+			
 			guiManager = process.StartGUIManager ();
 
 			//FIXME: conditionally add event handlers
@@ -625,9 +670,13 @@ namespace DebuggerServer
 		
 		void NotifyStarted ()
 		{
-			initializing = true;
-			controller.OnMainProcessCreated(process.ID);
-			initializing = false;
+			try {
+				initializing = true;
+				controller.NotifyStarted();
+				initializing = false;
+			} catch (Exception ex) {
+				Console.WriteLine (ex);
+			}
 		}
 
 		void OnTargetOutput (bool is_stderr, string text)
@@ -650,31 +699,13 @@ namespace DebuggerServer
 		void ResetTaskQueue ()
 		{
 			lock (debugger) {
-				internalInterruptionRequested = false;
 				stoppedWorkQueue.Clear ();
 			}
 		}
 		
-		void RunWhenStopped (ST.WaitCallback cb)
-		{
-			lock (debugger)
-			{
-				if (process.MainThread.IsStopped) {
-					cb (null);
-					return;
-				}
-				stoppedWorkQueue.Add (cb);
-				
-				if (!internalInterruptionRequested) {
-					internalInterruptionRequested = true;
-					process.MainThread.Stop ();
-				}
-			}
-		}
-		
 		void LogEvent (MD.TargetEventArgs args)
 		{
-			Console.WriteLine ("Server OnTargetEvent: {0} stopped:{1} data:{2} internal:{3} queue:{4} thread:{5} running:{6}", args.Type, args.IsStopped, args.Data, internalInterruptionRequested, stoppedWorkQueue.Count, args.Frame != null ? args.Frame.Thread : null, running);
+			Console.WriteLine ("Server OnTargetEvent: {0} stopped:{1} data:{2} queue:{3} thread:{4} running:{5}", args.Type, args.IsStopped, args.Data, stoppedWorkQueue.Count, args.Frame != null ? args.Frame.Thread : null, running);
 		}
 
 		private void OnTargetEvent (MD.Thread thread, MD.TargetEventArgs args)
@@ -696,13 +727,6 @@ namespace DebuggerServer
 				if (isStop) {
 					
 					lock (debugger) {
-						// The process was stopped, but not as a result of the internal stop request.
-						// Reset the internal request flag, in order to avoid the process to be
-						// automatically restarted
-						if (args.Type != MD.TargetEventType.TargetInterrupted && args.Type != MD.TargetEventType.TargetStopped)
-							internalInterruptionRequested = false;
-						
-						notifyToClient = notifyToClient && !internalInterruptionRequested;
 						
 						if (stoppedWorkQueue.Count > 0) {
 							// Execute queued work in another thread with a small delay
@@ -715,10 +739,6 @@ namespace DebuggerServer
 										cb (null);
 									}
 									stoppedWorkQueue.Clear ();
-									if (internalInterruptionRequested) {
-										internalInterruptionRequested = false;
-										resume = true;
-									}
 								}
 								if (resume)
 									guiManager.Continue (process.MainThread);
@@ -773,8 +793,8 @@ namespace DebuggerServer
 				}
 
 				if ((args.Type == MD.TargetEventType.UnhandledException || args.Type == MD.TargetEventType.Exception) && (args.Data is TargetAddress)) {
-					EvaluationContext ctx = new EvaluationContext (args.Frame.Thread, args.Frame, -1);
-					targetArgs.Exception = new LiteralValueReference (ctx, "Exception", args.Frame.ExceptionObject).CreateObjectValue ();
+					MdbEvaluationContext ctx = new MdbEvaluationContext (args.Frame.Thread, args.Frame, -1);
+					targetArgs.Exception = LiteralValueReference.CreateTargetObjectLiteral (ctx, "Exception", args.Frame.ExceptionObject).CreateObjectValue ();
 				}
 
 				running = false;
@@ -802,16 +822,14 @@ namespace DebuggerServer
 
 		public void CancelRuntimeInvokes ()
 		{
-			asyncEvaluationTracker.Stop ();
-			invokeManager.AbortAll ();
-			asyncEvaluationTracker.WaitForStopped ();
+			mdbObjectValueAdaptor.CancelAsyncOperations ();
 		}
 		
-		public void WaitRuntimeInvokes ()
+/*		public void WaitRuntimeInvokes ()
 		{
 			invokeManager.WaitForAll ();
 		}
-		
+*/		
 		private void OnProcessCreatedEvent (MD.Debugger debugger, MD.Process process)
 		{
 			WriteDebuggerOutput (string.Format ("Process {0} created.\n", process.ID));
@@ -839,12 +857,22 @@ namespace DebuggerServer
 		
 		private void OnModuleLoadedEvent (Module module)
 		{
-			WriteDebuggerOutput (string.Format ("Module {0} loaded.\n", module.Name));
+			SourceFile[] sfiles = module.Sources;
+			string[] files = new string [sfiles.Length];
+			for (int n=0; n<files.Length; n++)
+				files [n] = sfiles [n].FileName;
+			controller.NotifySourceFileLoaded (files);
+			WriteDebuggerOutput (string.Format ("Module loaded: {0}.\n", module.Name));
 		}
 
 		private void OnModuleUnLoadedEvent (Module module)
 		{
-			WriteDebuggerOutput (string.Format ("Module {0} unloaded.\n", module.Name));
+			SourceFile[] sfiles = module.Sources;
+			string[] files = new string [sfiles.Length];
+			for (int n=0; n<files.Length; n++)
+				files [n] = sfiles [n].FileName;
+			controller.NotifySourceFileUnloaded (files);
+			WriteDebuggerOutput (string.Format ("Module unloaded: {0}.\n", module.Name));
 		}
 
 		private void OnTargetExitedEvent (MD.Debugger debugger)
@@ -885,5 +913,30 @@ namespace DebuggerServer
 				}
 			}
 		}
+		
+		string[] ToArgsArray (string str)
+		{
+			List<string> args = new List<string> ();
+			bool escaping = false;
+			StringBuilder currentArg = new StringBuilder ();
+			for (int n=0; n<str.Length; n++) {
+				char c = str [n];
+				if (c == '\\' && n + 1 < str.Length) {
+					if (!escaping || (escaping && (str[n+1] == '"' || str[n+1] == '\\')))
+						c = str [++n];
+				} else if (c == '"') {
+					escaping = !escaping;
+					continue;
+				} else if (c == ' ' && !escaping) {
+					args.Add (currentArg.ToString ());
+					currentArg = new StringBuilder ();
+					continue;
+				}
+				currentArg.Append (c);
+			}
+			if (currentArg.Length > 0)
+				args.Add (currentArg.ToString ());
+			return args.ToArray ();
+		}
 	}
 }
